#!/usr/bin/perl

# $Id: mzg.pl,v 1.2 2005/05/11 00:08:50 vixie Exp $

use strict;
use Net::DNS;

my ($mz, $ns, $dn, @others) = @ARGV;
die "usage: $0 metazone nameserver dirname"
	unless length $mz && length $ns && length $dn;

die "please upgrade Net::DNS to version >= 0.45"
	unless Net::DNS->version >= 0.45;
my $res = Net::DNS::Resolver->new;

my $mz_version = 42;

(my $mzp = $mz) =~ s/(\W)/\\$1/g;
$mzp = '\.'.$mzp.'$';

our %servers = ( );
my @masters = ( );
my @allow_transfer = ( );
my @also_notify = ( );
my @zones = ( );
$res->nameservers($ns);
my @zone = $res->axfr($mz);
foreach my $rr (@zone) {
	my($attr, $pref, $name, $list, $zone);
	if ($rr->type eq 'NS' || $rr->type eq 'SOA') {
		if ($rr->name ne $mz) {
			badrr($rr);
		}
		if ($rr->type eq 'SOA' && $rr->minimum != $mz_version) {
			die "mz version (in SOA.MINIMUM) out of range";
		}
	} elsif ($rr->type eq 'A' || $rr->type eq 'AAAA') {
		if ($rr->name =~ /\.servers\.$mz$/) {
			($name = $rr->name) =~ tr/A-Z/a-z/;
			$servers{$name} = [ ] unless defined $servers{$name};
			push @{$servers{$name}}, $rr->address;
		} else {
			badrr($rr);
		}
	} elsif ($rr->type eq 'MG') {
		($attr = $rr->name) =~ s/\.$mz$//;
		$attr =~ tr/A-Z/a-z/;
		if ($attr eq 'allow-transfer') {
			push @allow_transfer, $rr->mgmname;
		} else {
			badrr($rr);
		}
	} elsif ($rr->type eq 'RT') {
		($attr = $rr->name) =~ s/\.$mz$//;
		$attr =~ tr/A-Z/a-z/;
		$pref = $rr->preference;
		($name = $rr->intermediate) =~ tr/A-Z/a-z/;
		$list = undef;
		if ($attr eq 'masters') {
			$list = \@masters;
		} elsif ($attr eq 'also-notify') {
			$list = \@also_notify;
		} else {
			badrr($rr);
		}
		push @$list, { pref => $pref, name => $name };
	} elsif ($rr->type eq 'PTR') {
		($zone = $rr->name) =~ s/\.zones\.$mz$//;
		$zone =~ tr/A-Z/a-z/;
		($name = $rr->ptrdname) =~ tr/A-Z/a-z/;
		if ($zone eq $name) {
			push @zones, $zone;
		} else {
			badrr($rr);
		}
	} else {
		badrr($rr);
	}
}

reorder(\@masters); resolve(\@masters);
reorder(\@also_notify); resolve(\@also_notify);
keyify(\@allow_transfer);

die "no master servers defined" unless $#masters >= $[;

foreach my $zone (@zones) {
	print "zone \"$zone\" {\n";
	print "\ttype slave;\n";
	printf "\tfile \"%s/%s\";\n", $dn, $zone;
	print "\tmasters { ", join('; ', @others, @masters, ''), "};\n";
	print "\tallow-transfer { ", join('; ', @allow_transfer, ''),
		"};\n" if $#allow_transfer >= $[;
	print "\talso-notify { ", join('; ', @others, @also_notify, ''),
		"};\n" if $#also_notify >= $[;
	print "};\n\n";
}

exit 0;

sub badrr { (my $rr) = @_; die "ungrokkable <".$rr->string.">"; }

sub reorder {
	(my $ref) = @_;

	my @new = ( );
	foreach my $el (sort { $a->{pref} <=> $b->{pref} } @$ref) {
		push @new, $el->{name};
	}
	@$ref = @new;
}

sub keyify {
	(my $ref) = @_;

	my @new = ( );
	foreach my $el (@$ref) {
		push @new, "key $el";
	}
	@$ref = @new;
}

sub resolve {
	(my $ref) = @_;

	my @new = ( );
	foreach my $el (@$ref) {
		die "no A or AAAA for $el" unless defined $servers{$el};
		push @new, grep /:/, @{$servers{$el}};	# IPv6 first
		push @new, grep !/:/, @{$servers{$el}};	# non-IPv6 last
	}
	@$ref = @new;
}

